/******************************************************************************/
/*									      */
/*	      Linear Prediction based DTMF receiver for AT&T WE (R) DSP32     */
/*									      */
/******************************************************************************/
/*									      */
/*			COPYRIGHT (c) 1987 AT&T				      */
/*									      */
/******************************************************************************/
/*									      */
/*	FILE = madtmf3.s						      */
/*									      */
/*	This is version 3 of the linear prediction (LP) based dtmf receiver.  */
/*									      */
/*	The following code implements 32 LP based DTMF receivers    	      */
/*	on the AT&T WE (R) DSP32 operating with a 160nsec instruction rate.   */
/*	The program is stored in ROM in bank 0 and the program state var's.   */
/*	are stored in bank1 RAM (memory mode 0). During the 250usec sample    */
/*	period, there is only time to execute an additional 36 instructions.  */
/*									      */
/*	The DSP32 receives sampled data on the serial input port in mu-law    */
/*	format using passive clocks. DMA is used to buffer the sampled data   */
/*	from each codec. An external sync pulse tells the AT&T WE (R) DSP32   */
/*	that codec 1 is beginning to transmit samples.  	              */
/*	The sync pulse should occur between the load of the 32nd codec and    */
/*	the 1st codec every other time. It should go high immediately after   */
/*	codec 32's sync resets and go low at least 1.92us before codec 1      */
/*	is synced.							      */
/*									      */
/*	When a digit has been validated, it is written to pdr.  The pdr       */
/*	contains the detected digit in the lower 4 bits and the receiver      */
/*	number in bits 5-9.						      */
/*									      */
/******************************************************************************/
/*									      */
/*			DISCLAIMER AND NOTICE				      */
/*									      */
/*	AT&T and its SUBSIDIARIES (hereinafter AT&T) reserve the right to     */
/*	make changes in this APPLICATION NOTE and the semiconductor devices   */
/*	or device specifications to which it is directed.		      */
/*									      */
/*	AT&T makes no representations or warranties, expressly or implied.    */
/*	By way of example but not of limitation, AT&T makes no representa-    */
/*	tions or warranties of merchantability or fitness for any particular  */
/*	purpose, or that the use of the INFORMATION in this APPLICATION NOTE  */
/*	or any of it will not infringe any patent, copyright or other intel-  */
/*	lectual property right of a third party.  AT&T shall not be held to   */
/*	any liability with respect to any claim by a USER or any third party  */
/*	on account of, or arising from the use of the INFORMATION or any      */
/*	part of it.  							      */
/*									      */
/*	AT&T further does not grant any license, express or implied, under    */
/*	any patent, copyright or other intellectual property right of AT&T    */
/*	covering or relating to any process, machine or combination in which  */
/*	the INFORMATION and/or semiconductor device or devices may be or are  */
/*	used.								      */
/*									      */
/*	The software code and other information contained in this APPLICA-    */
/*	TION NOTE are copyrighted and all rights are reserved by AT&T.  The   */
/*	software code is intended for use only in an AT&T Digital Signal      */
/*	Processor.  No other use is authorized.				      */
/*									      */
/******************************************************************************/
/*.rsect ".rom"*/

	pin=inbuf	/* set input DMA pointer to beginning of input buffer */
	ioc=0x2040	/* set input serial i/o for 8-bit serial DMA mode     */
	dauc=0x0

wait:	if (syc) goto wait	/* wait for channel 0			      */
	nop

	pin=inbuf	/* set input DMA pointer to beginning of input buffer */

	r15=8		/* used for two floating point location post-increment*/

	r17=-60		/* used for post-decrement the BPF coeffs. pointer    */
			/* back to the beginning of the coeffs after filtering*/

/*****************************************************************************/
/*									     */
/*	Zero out the memory locations of the low band energies, the auto-    */
/*	correlation values and the state variables for their calculation     */
/*	for all 32 receivers.						     */
/*									     */
/*****************************************************************************/
	r1=shortel
	r2=zero1
	r3=30
initloop:*r1++=a0=*r2
	*r1++=a0=*r2
	*r1++=a0=*r2
	*r1++=a0=*r2
	*r1++=a0=*r2
	*r1++=a0=*r2
	*r1++=a0=*r2
	*r1++=a0=*r2
	*r1++=a0=*r2
	*r1++=a0=*r2
	if (r3-- >= 0) goto initloop
	*r1++=a0=*r2
	*r1++=a0=*r2
	*r1++=a0=*r2

/*****************************************************************************/

begin:	r13=0			/* set channel counter register to zero     */

	r18=21-2		/* initialize verification hold counter     */

	r9=autoc11l		/* initialize pointer to correlation state  */
				/* variables				    */

iwait:	r1=inbuf		/* r1 is the inbuf pointer		   */
	r4=bufio		/* r4 is the working buffer, bufio, pointer*/
	r16=4			/* set the post-incrementer to four bytes  */

	pin-ilimit
iwait2:	if(le)goto iwait2    /* wait until serial input DMA has filled inbuf */
	pin-ilimit

	pin=inbuf	     /* reset the input serial DMA pointer to inbuf  */

	*r4++=a0=ic(*r1++r16) /* Transfer the first 32 of the 64 samples in */
	*r4++=a0=ic(*r1++r16) /* inbuf to bufio. This effecively subsamples */
	*r4++=a0=ic(*r1++r16) /* the 8 kHz sampled input down to a 4 kHz    */
	*r4++=a0=ic(*r1++r16) /* sample rate.  In the process, convert from */
	*r4++=a0=ic(*r1++r16) /* mu-law to float format.		    */
	*r4++=a0=ic(*r1++r16)
	*r4++=a0=ic(*r1++r16)
	*r4++=a0=ic(*r1++r16)
	*r4++=a0=ic(*r1++r16)
	*r4++=a0=ic(*r1++r16)
	*r4++=a0=ic(*r1++r16)
	*r4++=a0=ic(*r1++r16)
	*r4++=a0=ic(*r1++r16)
	*r4++=a0=ic(*r1++r16)
	*r4++=a0=ic(*r1++r16)
	*r4++=a0=ic(*r1++r16)
	*r4++=a0=ic(*r1++r16)
	*r4++=a0=ic(*r1++r16)
	*r4++=a0=ic(*r1++r16)
	*r4++=a0=ic(*r1++r16)
	*r4++=a0=ic(*r1++r16)
	*r4++=a0=ic(*r1++r16)
	*r4++=a0=ic(*r1++r16)
	*r4++=a0=ic(*r1++r16)
	*r4++=a0=ic(*r1++r16)
	*r4++=a0=ic(*r1++r16)
	*r4++=a0=ic(*r1++r16)
	*r4++=a0=ic(*r1++r16)
	*r4++=a0=ic(*r1++r16)
	*r4++=a0=ic(*r1++r16)
	*r4++=a0=ic(*r1++r16)
	*r4++=a0=ic(*r1++r16)

	/* for the next channel to undergo verification tests let:	     */
	r5=autoc22l		/*					     */
	*r5++=a0=*r9++		/* r(2,2)=r(1,1) in the low band	     */
	*r5++=a0=*r9++r15	/* r(1,2)=r(0,1) in the low band	     */
	*r5++=a0=*r9++		/* r(2,2)=r(1,1) in the high band	     */
	*r5=a0=*r9++r15		/* r(1,2)=r(0,1) in the high band	     */

	r9=r9-24		/* restore r9 to point to r(1,1) of the next */
				/* channel to undergo verification tests     */


/*****************************************************************************/
/*									     */
/*	Compute the bandpass filters and autocorrelation matrices for 32     */
/*	receivers.  Memory for each receiver is stored sequentially starting */
/*	at the locations shown below.					     */
/*									     */
/*****************************************************************************/
	r12=32-2
	r11=bufio		/* 4KHz sampled input location */
	r5=lambda		/* location of the forgetting factor */
	r6=outl			/* location of the bandpass filter outputs */
	r8=y2l			/* low band y(n-2)  */
	r10=shortel		/* start loc. of low band energy for recvrs. */
	r2=lcoef		/* loc. of BPF coefficients */
	r3=lostate		/* start loc. of state variables for recvrs. */
	r4=r3
	r1=autoc20l		/* start loc. of autocorrelation recvrs vars. */
	r7=r6
	  
/*			Low bandpass filter			*/

moreits:a0 = *r11 * *r2++		      /* input * constant multiplier */
 	*r4++ = a0 = a0 - *r3 * *r2++         /* d[i,1] term */
	a0 = a0 + *r3++ * *r2++   	      /* n[i,1] term */

 	a0 = a0 - *r3++ * *r2++               /* d[i,1] term */
	*r4++ =   a0 = a0 - *r3-- * *r2++     /* d[i,2] term */
	a0 = a0 + (*r4++ = *r3++) * *r2++     /* n[i,1] term */
	*r6++ = a0 = a0 + *r3++ * *r2++       /* n[i,2] term */
	  
/*			High bandpass filter			*/

	a0 = *r11++ * *r2++    		      /* input * constant multiplier */
 	a0 = a0 - *r3++ * *r2++               /* d[i,1] term */
	*r4++ =   a0 = a0 - *r3-- * *r2++     /* d[i,2] term */
	a0 = a0 + (*r4++ = *r3++) * *r2++     /* n[i,1] term */
	a0 = a0 + *r3++ * *r2++		      /* n[i,2] term */

 	a0 = a0 - *r3++ * *r2++               /* d[i,1] term */
	*r4++ =   a0 = a0 - *r3-- * *r2++     /* d[i,2] term */
	a0 = a0 + (*r4++ = *r3++) * *r2++     /* n[i,1] term */
	*r6-- =a0= a0 + *r3++ * *r2++r17      /* n[i,2] term */

/****************************************************************************/
/*									    */
/*	This section adaptively computes the correlation values for both    */
/*	low and high bands.						    */
/*									    */
/****************************************************************************/

/*	for the low band, update the LBE, r(0,2), r(1,1), r(0,1)              */

	a0=*r7**r8++		/* a0 = y(n-2) * y(n) 			      */
	*r1++=a0=a0+*r1**r5	/* r(0,2) = lambda * r(0,2) + y(n-2) * y(n)   */
	a0=(*r8++=*r8--)**r8	/* a0 = y(n-1) * y(n-1), y(n-2) = y(n-1)      */
	*r10++=a1=a0+*r10	/* LBE update				      */
	*r1++=a0=a0+*r1**r5	/* r(1,1) = lambda * r(1,1) + y(n-1) * y(n-1) */
	a0=(*r8++=*r7++)**r8	/* a0 = y(n-1) * y(n), y(n-1) = y(n)          */
	*r1++=a0=a0+*r1**r5	/* r(0,1) = lambda * r(0,1) + y(n-1) * y(n)   */

/*	for the high band, update r(0,2), r(1,1), r(0,1)	              */

	a0=*r7**r8++		/* a0 = y(n-2) * y(n)			      */
	*r1++=a0=a0+*r1**r5	/* r(0,2) = lambda * r(0,2) + y(n-2) * y(n)   */
	a0=(*r8++=*r8--)**r8	/* a0 = y(n-1) * y(n-1), y(n-2) = y(n-1)      */
	*r1++=a0=a0+*r1**r5	/* r(1,1) = lambda * r(1,1) + y(n-1) * y(n-1) */
	a0=(*r8++=*r7--)**r8	/* a0 = y(n-1) * y(n), y(n-1) = y(n)          */
loop:	if(r12-->=0)goto moreits	/* do this for all 32 channels        */
	*r1++=a0=a0+*r1**r5	/* r(0,1) = lambda * r(0,1) + y(n-1) * y(n)   */


/*****************************************************************************/
/*									     */
/*	Compute the magnitude and phase of each tone.			     */
/*									     */
/*****************************************************************************/
	if (r18-- >= 0) goto iwait

	r2=tempint		/* points to a scratch memory location for    */
				/* data type conversions		      */

	r9=r9+24		/* r9 now points to r(1,1) of next channel to */
				/* undergo verification tests		      */

	r3=r13*2		/* r3 is integer (2 bytes) channel offset     */

	r5=aperture+r3
	r1=*r5			/* r1 contains aperture state of this channel */
	nop				/*nop needed for latent DAU COND test */
	if(eq)goto wide1		/* test aperture size */
	r5=twenty4			/* r5 points to 24    */

	r10=hitblt		/* aperture was narrow set pointers to tables */
	goto cont1		/* for narrow angle aperture tests	      */
	r12=lotblt

wide1:	r12=lotbl3		/* aperture was wide set pointers to tables   */
	r10=hitbl3		/* for wide angle aperture tests	      */

cont1:	r4=r3*2			/* r4 is float (4 bytes) channel offset       */
	r8=shortel+r4		/* r8 points to LBE estimate for this channel */
	*r2=r13
	a1=float(*r2)			/* convert n=receiver no. to float    */
	r1=threshs
	a0 = *r8 - *r1			/* LBE thresh test for tone presence  */
	a1=a1**r5			/* a1=n*24 */
	*r2=a1=int(a1)			/* convert a1=n*24 back to int        */
	nop
	if (ale) goto notone		/* if LBE < 265,000 goto notone       */
	r16h=*r2		/* r16h=chn# NOT int(a1) due to latency	      */

	r4=*r2				/* r4=n*24 */
	r7=autoc22l
	call root (r14)		/* do mag. & angle tests for low band pole    */
	r8=autoc20l+r4
	r12=r10
	r8=autoc20h+r4
	r7=autoc22h
	call root (r14)		/* do mag. & angle tests for high band pole  */
	r4=*r6				/* r4 = row location of digit        */
	 
/*****************************************************************************/
/*									     */
/*	Decode digits and test for valid tone.				     */
/*									     */
/*****************************************************************************/
	r8=r3*2			/* r8 is float (4 bytes) channel offset	     */
	r2=aperture+r3		/* points to this channel's aperture state   */
	r4=r4*2
	if (lt) goto notone		/* if r4=-1, there is no touch tone  */
	r7=*r6				/* r7 = column location of digit     */
	r4=r4*2
	r7=r7*2
	if (lt) goto notone		/* if r7=-1, there is no touch tone  */
	r4=r4*2
	r7=r7+r4
	r7=r7+_romtbl
	r10=*r7				/* r10 contains the decoded digit    */

	r4=_ttdigit+r8
	r4=r4+2
	r5=*r4--
	r6=*r4
	r5=r5-r10
	if (eq) goto tone1		/* if D(k) = D(k-2) this is old digit*/
	r1=1
					/* if not then it may be a new one   */
	*r2=r1				/* set aperture for narrow window */
	r6=r6-r10
	if (ne) goto tone1		/* if D(k) not = D(k-1) need one more */
					/* frame of data		      */

	r16=r16^r10		/* r16h has channel no. r16l has DTMF digit   */
	goto tone1
	pdr=r16			/* send channel no. and detected digit to uP  */

notone:	r10=-1			/* some verification test failed	      */
	r2=aperture+r3
	r1=0
	*r2=r1				/* reset aperture to wide width       */
tone1:	r8=r3*2
	r1=_ttdigit+r8
	r7=*r1
	*r1++=r10			/* update past digit D(k-1)           */
	*r1=r7				/* update past digit D(k-2)           */

	r4=r3*2				/* r4 is float (4 bytes) chan offset  */
	r1=shortel+r4
	r2=zero1
	*r1=a0=*r2			/* zero out LBE for this channel      */
	r4=tempint
	r1=*r4
	r5=aperture+r3
	r6=*r5				/* this is TESTED in CAU COND TEST    */
	r1=*r4			/* this is NOT TESTED due to latent read      */
	if (gt) goto rett	/* if aperture=0, zero out this chan's 10     */
	r1=autoc20l+r1		/* autocorrelation variables.	  	      */
zeroize:*r1++=a0=*r2
	*r1++=a0=*r2
	*r1++=a0=*r2
	*r1++=a0=*r2
	*r1++=a0=*r2
	*r1++=a0=*r2
rett:	r13-31
	if (eq) goto begin	/* if that was chan 31 goto begin 	     */
	nop
	goto iwait
	r13=r13+1		/* increment channel counter register	     */

/*--------------------------------------------------------------------------*/
/*	subroutine root:  Given the autocorrelation matrix R and the cross-
	correlation vector r, this subroutine will determine and test the
	magnitude of the tone and determine the angle (the row or column loc.
	of the digit ) of the tone.  It expects r8 to point to the beginning
	of R and r12 to point to the correct angle decision table. The integer
	result ( row or column loc. of the digit ) is written to the location
	called tone. If the magnitude test fails control vectors to location,
	notone.  If angle tests fail the output is -1.
*/
root:	a2=-*r8**r8++		/* a2 = r(1,1) * r(0,2)			  */
	a3=*r8--**r7++		/* a3 = r(2,2) * r(1,1)			  */
	a1=*r8++r15**r7--	/* a1 = r(0,2) * r(1,2) 		  */
	a1=a1-*r8**r7++		/* a1 = r(0,2) * r(1,2) - r(2,2) * r(0,1) */
	a3=a3-*r7**r7		/* a3 = r(2,2) * r(1,1) - r(1,2) * r(1,2) */
	a2=a2+*r8++**r7++	/* a2 = r(1,2) * r(0,1) - r(1,1) * r(0,2) */
	r6=thresh		/* points to pole mag threshold		  */
	a0 = a2 - a3 * *r6	/* a0 = a2 - threshold * det 		  */
	r1=four			/* points to floating point four	  */
	a2 = a2 * a3		/* a2 = a2 * det 			  */
	r2=det
	if(ale)goto notone	/* if magnitude < thresh goto notone 	  */
	r11=minus1
	*r2 = a1 = a1 * a1	/* a1 = (a1 * det)^2			  */
	a2 = a2 * *r1		/* a2 = 4 * (a2 * det)			  */
	r1=2
	a0=*r11++
whchton:a3 = a2 - *r2 * *r12++		/* a3=4*a2*det - a1^2 * anglethresh++ */
	a1 = a2 - *r2 * *r12++		/* a3=4*a2*det - a1^2 * anglethresh++ */
	nop
	r6=tone
	a1=a3*a1		/* a1 is negitive if angle is within aperture */
	if(r1-->=0)goto whchton
	a0=ifalt(*r11++)
	return (r14)
	*r6 = a0 = int(a0)	/* output is in mem location - tone	      */
				/* if no tone found output is -1, 	      */
				/* else output is one of 0,1,2,3 -- the i'th  */
				/* tone of the band being analyzed	      */
	
/*****************************************************************************/
/*									     */
/*			PROGRAM COEFFICIENTS				     */
/*									     */
lambda:	float 0.981481481
threshs:float 265000.		/* Low band energy threshold */
thresh: float .95		/* pole magnitude threshold .9814814 */
minus1:	float -1.
zero:	float 0.
one:	float 1.
two:	float 2.0
three:	float 3.0
four:	float 4.0
twenty4:float 24.0
_romtbl:int 1, 2, 3, 10
	int 4, 5, 6, 11
	int 7, 8, 9, 12
	int 14, 0, 15, 13

/*	Low bandpass filter coefficients				     */
lcoef:	float .160742086
	float -.527429, 1.0
	float -.189495, .79558, .830501449, .999648679

/*	High bandpass filter coefficients				     */
hcoef:	float .045015823
	float 1.202186, .557487, -1.470971950, .998564553
	float .660893, .877532, -.304368084, .999705344

/*****************************************************************************/
/*			Angle threshold coefficients			     */
/*									     */
/*			angle = freq. * 180deg. / 2000hz		     */
/*	low band angle computation:  Table value = tan^2(angle) + 1	     */
/*	high band angle computation: Table value = tan^2(180 - angle) + 1    */
/*****************************************************************************/
/*		  Table value	freg(hz) freg offset(%)			     */
/*									     */
/* Wide aperture:							     */
lotbl3:	float	    3.97154     /* 665	-4.6 */
	float	    6.13336    	/* 735	 5.4 */
	float	    6.11845     /* 735	-4.5 */
	float	    11.8574    	/* 812	 5.4 */
	float	    11.7947     /* 812	-4.7 */
	float	    38.8135    	/* 897   5.3 */
	float	    37.6823     /* 895	-4.9 */
	float	    4156.93     /* 990	 5.2 */
hitbl3: float	    6.02902    	/* 1267   4.8 */
	float	    19.2554     /* 1146  -5.2 */
	float	    2.91950    	/* 1398	  4.6 */
	float	    6.12220    	/* 1265  -5.3 */
	float	   1.738967    	/* 1548   4.8 */
	float	    2.88772    	/* 1401  -5.1 */
	float	   1.243161    	/* 1708   4.6 */
	float	   1.749901    	/* 1546  -5.3 */

/* Narrow aperture:							     */
lotblt:	float     4.38518    	/* 683  -2.0 */
	float     5.31677    	/* 714   2.4 */
	float     6.93326	/* 752	-2.3 */
	float     9.33380    	/* 788   2.3 */
	float     14.69761	/* 832  -2.3 */
	float     23.9671    	/* 869   1.9 */
	float     55.7752	/* 914  -2.8 */
	float     322.374    	/* 965   2.6 */
hitblt: float     7.08055    	/* 1245   2.9 */
	float     12.441	/* 1183  -2.1 */
	float     3.32711    	/* 1369   2.5 */
	float     4.59431	/* 1309  -2.0 */
	float     1.916202    	/* 1514   2.5 */
	float     2.397		/* 1447  -2.0 */
	float     1.312574    	/* 1675   2.6 */
endpgm:	float     1.532391    	/* 1599  -2.1 */

/*****************************************************************************/
/*									     */
/*			PROGRAM STATE VARIABLES				     */
/*									     */
.rsect ".hi_ram"
zero1:	float 0.
outl:	float 0.0		/* output from low bandpass filter */
outh:	float 0.0		/* output from high bandpass filter */
lostate:224*float 0.		/* Begin bandpass filter state variables */
det:	float 0.		/* (a1*det)^2 */
aperture:32*int 0		/* contains present aperture size per recvr */
tempint:int 0			/* 24*n */
tone:	int 0			/* present low or high band tone detected */
_ttdigit:64*int -2		/* contains past 2 digits detected per recvr */

/*****************************************************************************/
/*									     */
/*	Low band state variables (frame by frame basis). 		     */
/*									     */
/*****************************************************************************/
shortel:32*float 0.	/* energy in low band for one frame per recvr */

autoc22l:  float 0.	/* correlations for current channel */
autoc12l:  float 0.
autoc22h:  float 0.
autoc12h:  float 0.

autoc20l:  float 0.	/* correlation values for all receivers */
autoc11l:  float 0.
autoc10l:  float 0.
autoc20h:  float 0.
autoc11h:  float 0.
autoc10h:  187*float 0.

y2l:	float 0.	/* delayed input samples for all receivers */
y1l:	float 0.
y2h:	float 0.
y1h:	125*float 0.

/*****************************************************************************/
/*				DMA location				     */
/*****************************************************************************/
inbuf:	63*float 0.
ilimit:	float 0.
bufio:	32*float 0.
/******************************************************************************/
