/******************************************************************************/
/*									      */
/*		DFT based DTMF receiver for AT&T WE (R) DSP32		      */
/*									      */
/******************************************************************************/
/******************************************************************************/
/*									      */
/*			COPYRIGHT (c) 1987 AT&T				      */
/*									      */
/******************************************************************************/
/******************************************************************************/
/*									      */
/*	FILE = sadtmf12.s						      */
/*									      */
/*	This is version 2 of the adaptive dtmf receiver.  It contains the     */
/*	aperture adjustment that was accidentally omitted in version 1.       */
/*									      */
/*	The following code implements 1 Adaptive RLS based DTMF receiver on   */
/*	the AT&T WE (R) DSP32. The program and its coefficients are stored    */
/*	in bank 0 RAM and the state variables in bank 1 RAM.  		      */
/*									      */
/*	The DSP32 receives sampled data on the serial input port in mu-law    */
/*	format using passive clocks.  DMA is used to buffer an entire 8KHz    */
/*	sample frame (106 samples to be converted to 4KHz sample rate).       */
/*	When a valid digit is detected, it is written to the memory location  */
/*	called output. All following valid detected digits are appended to    */
/*	this location.
/*									      */
/*	Thirty two DTMF receivers can be implemented on 1 AT&T WE (R) DSP32   */
/*	running at 160 nsec when ROM is used to store the program and its     */
/*	coefficients and RAM is used to store the state variables.  Memory    */
/*	mode 0 must be used in this case.				      */
/*									      */
/******************************************************************************/
/******************************************************************************/
/*									      */
/*			DISCLAIMER AND NOTICE				      */
/*									      */
/*	AT&T and its SUBSIDIARIES (hereinafter AT&T) reserve the right to     */
/*	make changes in this APPLICATION NOTE and the semiconductor devices   */
/*	or device specifications to which it is directed.		      */
/*									      */
/*	AT&T makes no representations or warranties, expressly or implied.    */
/*	By way of example but not of limitation, AT&T makes no representa-    */
/*	tions or warranties of merchantability or fitness for any particular  */
/*	purpose, or that the use of the INFORMATION in this APPLICATION NOTE  */
/*	or any of it will not infringe any patent, copyright or other intel-  */
/*	lectual property right of a third party.  AT&T shall not be held to   */
/*	any liability with respect to any claim by a USER or any third party  */
/*	on account of, or arising from the use of the INFORMATION or any      */
/*	part of it.  							      */
/*									      */
/*	AT&T further does not grant any license, express or implied, under    */
/*	any patent, copyright or other intellectual property right of AT&T    */
/*	covering or relating to any process, machine or combination in which  */
/*	the INFORMATION and/or semiconductor device or devices may be or are  */
/*	used.								      */
/*									      */
/*	The software code and other information contained in this APPLICA-    */
/*	TION NOTE are copyrighted and all rights are reserved by AT&T.  The   */
/*	software code is intended for use only in an AT&T Digital Signal      */
/*	Processor.  No other use is authorized.				      */
/*									      */
/******************************************************************************/
.global	output,aperture,autoc22l,autoc22,lambda,pause,_ttdigit,contt
	pin=inbuf
	dauc=0x0
	ioc=0x2547
	r18=24				/*  26 - 2 frames  */
	r13=output
/*****************************************************************************/
/*									     */
/*	Wait for DMA inbuf full then transfer every other sample to bufio    */
/*	(4KHz sample rate).						     */
/*									     */
/*****************************************************************************/
iwait:	r1=inbuf
	r3=53-3
	r4=bufio
	r16=8
	pin-ilimit
iwait2:	if(le)goto iwait2
	pin-ilimit

	*r4++=a0=ic(*r1++r16)  		 /* load bufio at 4KHz rate */
	pin=inbuf
ioxfr:	if(r3-->=0)goto ioxfr
	*r4++=a0=ic(*r1++r16)  		 /* load bufio at 4KHz rate */

	r1=bufio
	r2=bufpnt		/* *bufpnt contains starting address of bufio */
	*r2=r1

tmplop:	r1=shortel		/* energy in low band of one frame */
	r2=autoc22		/* energy in high band of one frame */
	r4=zero
	*r1++=a0=*r4
	r5=aperture
	r6=*r5 			/* r6 contains present aperture size */
	nop			/* nop needed for latent CAU COND test */
	if(gt)goto moreit1
	r3=7-2
/*****************************************************************************/
/*									     */
/*	If aperture=0.0 then zero out autocorrelation matrices.		     */
/*									     */
/*****************************************************************************/
zeroize:*r1++=a0=*r4
	if(r3-->=0)goto zeroize
	*r2++=a0=*r4
/*****************************************************************************/
/*									     */
/*	Compute the bandpass filters and autocorrelation matrices for the    */
/*	frame of data.							     */
/*									     */
/*****************************************************************************/
moreit1:r12=53-2		/* 53 sample frame */
	r2=bufpnt
	r11=*r2			/* r11 contains the address of bufio */
moreits:r6=outl			/* location of lpf output */
	a3 = *r11++		/* present sample */	
	r2=lcoef		/* location of lpf coefficients */
	r3=lostate		/* location of lpf state variables */
	r4=r3
	  
/*			Low bandpass filter			*/

        a0 = a3 * *r2++			      /* input * constant multiplier */
 	a0 = a0 - *r3++ * *r2++               /* d[i,1] term */
        *r4++ =   a0 = a0 - *r3-- * *r2++     /* d[i,2] term */
        a0 = a0 + (*r4++ = *r3++) * *r2++     /* n[i,1] term */
        a0 = a0 + *r3++ * *r2++		      /* n[i,2] term */

 	a0 = a0 - *r3++ * *r2++               /* d[i,1] term */
        *r4++ =   a0 = a0 - *r3-- * *r2++     /* d[i,2] term */
        a0 = a0 + (*r4++ = *r3++) * *r2++     /* n[i,1] term */
        *r6++ = a0 = a0 + *r3++ * *r2++       /* n[i,2] term */
	  
/*			High bandpass filter			*/

        a0 = a3 * *r2++    		      /* input * constant multiplier */
 	a0 = a0 - *r3++ * *r2++               /* d[i,1] term */
        *r4++ =   a0 = a0 - *r3-- * *r2++     /* d[i,2] term */
        a0 = a0 + (*r4++ = *r3++) * *r2++     /* n[i,1] term */
        a0 = a0 + *r3++ * *r2++               /* n[i,2] term */

 	a0 = a0 - *r3++ * *r2++               /* d[i,1] term */
        *r4++ =   a0 = a0 - *r3-- * *r2++     /* d[i,2] term */
        a0 = a0 + (*r4++ = *r3++) * *r2++     /* n[i,1] term */
        *r6++ = a0 = a0 + *r3++ * *r2++       /* n[i,2] term */

/****************************************************************************/
/*									    */
/*	This section adaptively computes the autocorrelation matrix and     */
/*	the cross-correlation vector for each frequency band according to   */
/*	the following equations: 					    */
/*									    */
/*	R(n) = lambda * R(n) + [y(n-1) y(n-2)] * [y(n-1) y(n-2)]^T 	    */
/*									    */
/*		r(n) = lambda * r(n) + yn * [y(n-1) y(n-2)]^T		    */
/*									    */
/****************************************************************************/
	r1=outl			/* location of lpf output */
	a1=*r1			/* low bandpass filtered sample yn  */
rls1:	r1=shortel		/* low band energy 		*/
	r2=crcor1l		/* low band cross-correlation element R(0,1) */
	r3=y2l			/* low band y(n-2)  */
	r5=lambda		/* forgetting factor */

/*			update the autocorrelation matrix R	            */

	a0=*r3++**r3		/* a0 = y(n-2) * y(n-2) */
	*r1++=a2=a0+*r1		/* Low band energy(LBE) = LBE * y(n-2)^2 */
	nop
	*r1++=a0=a0+*r1**r5	/* R(2,2) = lambda * R(2,2) + y(n-2)^2	 */	
	*r1++=a0=*r2		/* R(1,2) = R(0,1) */
	a0=*r3--**r3		/* a0 = y(n-1) * y(n-1) */
	*r1++=a0=a0+*r1**r5	/* R(1,1) = lambda * R(1,1) + y(n-1)^2   */

/*			update cross-correlation vector r		    */

	a0=(*r3=a1)**r3++	/* y(n-1) = yn & a0 = yn * y(n-2) */
	*r1++=a0=a0+*r1**r5	/* R(0,2) = lambda * R(0,2) + yn * y(n-2) */
	a0=(*r3++=*r3--)*a1	/* a0 = yn * y(n-1) & y(n-2) = y(n-1)     */
	*r1++=a0=a0+*r1**r5	/* R(0,1) = lambda * R(0,1) + yn * y(n-1) */

	r2=crcor1	    	/* high band cross-corr. element R(0,1) */
	r3=y2			/* high band y(n-2)  */
	r1=outh			/* location of hpf output */
	a1=*r1			/* high bandpass filtered sample yn  */
rls2:	r1=autoc22		/* high band energy  */

/*			update the autocorrelation matrix R		    */

	a0=*r3++**r3		/* a0 = y(n-2) * y(n-2) */
	*r1++=a0=a0+*r1**r5	/* R(2,2) = lambda * R(2,2) + y(n-2)^2	 */	
	*r1++=a0=*r2		/* R(1,2) = R(0,1) */
	a0=*r3--**r3		/* a0 = y(n-1) * y(n-1) */
	*r1++=a0=a0+*r1**r5	/* R(1,1) = lambda * R(1,1) + y(n-1)^2   */

/*			update cross-correlation vector r	   	    */

	a0=(*r3=a1)**r3++	/* y(n-1) = yn & a0 = yn * y(n-2) */
	*r1++=a0=a0+*r1**r5	/* R(0,2) = lambda * R(0,2) + yn * y(n-2) */
	a0=(*r3++=*r3--)*a1	/* a0 = yn * y(n-1) & y(n-2) = y(n-1)     */
	if(r12-->=0)goto moreits
	*r1++=a0=a0+*r1**r5	/* R(0,1) = lambda * R(0,1) + yn * y(n-1) */



/*****************************************************************************/
/*									     */
/*	Compute the magnitude and phase of each tone.			     */
/*									     */
/*****************************************************************************/
	r5=aperture
	r1=*r5
	nop			/* nop needed for latent DAU COND test */
	if(eq)goto wide1	/* test aperture size */
	nop
	r10=hitblt
	goto cont1
	r12=lotblt
wide1:	r12=lotbl3
	r10=hitbl3
cont1:	r8=shortel
	r1=threshs
	a0 = *r8++ - *r1
	3*nop
	if (ale) goto notone		/* if LBE < 265,000 goto notone */
	nop
	call root (r14)			/* compute mag. & phase for low tone */
	r8=autoc22l
	r12=r10
	r8=autoc22
	call root (r14)			/* compute mag. & phase for high tone */
	r4=*r6				/* store the row location of TT in r4 */
	 
/*****************************************************************************/
/*									     */
/*	Decode digits and test for valid tone.				     */
/*									     */
/*****************************************************************************/
	r2=aperture			/* preset r2=aperture, see below  */
	r4=r4*2
	if (lt) goto notone		/* if r4=-1, there is no touch tone */
	r3=*r6				/* store the col. location in r3 */
	r4=r4*2
	r3=r3*2
	if (lt) goto notone		/* if r3=-1, there is no touch tone */
	r4=r4*2
	r3=r3+r4
	r3=r3+_romtbl
	r10=*r3				/* r10 contains the decoded digit */
	r4=_ttdigit+2
	r5=*r4--
	r6=*r4
	r5=r5-r10
	if (eq) goto tone1		/* D(k) = D(k-2) ? */
	r1=1
	*r2=r1				/* set aperture for narrow window */
	r6=r6-r10
	if (ne) goto tone1		/* D(k) = D(k-1) ? */
	nop
	*r13++=r10			/* write valid digit to memory */
	goto tone1
	nop
notone:	r10=-1
	r2=aperture
	r1=0
	*r2=r1				/* set aperture for large window */
tone1:	r1=_ttdigit
	r3=*r1
	*r1++=r10			/* update past digit D(k-1) */
	*r1=r3				/* update past digit D(k-2) */
	if (r18-- >= 0) goto contt
	nop
pause:	nop

contt:	goto iwait
	nop

/*--------------------------------------------------------------------------*/
/*	subroutine root:  Given the autocorrelation matrix R and the cross-
	correlation vector r, this subroutine will determine and test the 
	magnitude of the tone and determine the phase (the row or column loc.
	of the digit) of the tone. It expects r8 to point to the beginning
	of R and r12 to point to the correct angle decision table. The integer
	result (row and column location of the digit) is written to the 
	location called tone.
*/

root:	r15=8
	r6=thresh
	a0=*r8--**r8++r15	/* a0 = R(2,2) * R(1,1) */
	a3=a0-*r8**r8		/* a3 contains determinant */
	r7=r8
	r7=r7+8
	a1= *r8--**r7++		/* a1 = R(1,2) * R(0,2) */
	a1=a1-*r8++**r7		/* a1 = R(1,2) * R(0,2) - R(2,2) * R(0,1) */
	a2= *r8++**r7--		/* a2 = R(1,2) * R(0,1) */
	a2=a2-*r8**r7		/* a2 = R(1,2) * R(0,1) - R(1,1) * R(0,2) */
	a0 = a2 - a3 * *r6	/* a0 = a2 - thresh(.95) * det */
	r1=four
	a2 = a2 * a3		/* a2 = a2 * det */
	r2=det
	if(ale)goto notone	/* 4 instruction latent DAU cond test from */
	nop			/* instruction a0 = a2 - *r6 * a3  	   */
	*r2 = a1 = a1 * a1	/* a1 = (a1 * det)^2 */
	a2 = a2 * *r1		/* a2 = 4 * (a2 * det) */
	r11=minus1
	r6=tone
	r1=2
	a0=*r11++
whchton:a3 = a2 - *r2 * *r12++		/* a3=4*a2*det - a1^2 * anglethresh++ */
	a1 = a2 - *r2 * *r12++		/* a3=4*a2*det - a1^2 * anglethresh++ */
	2*nop
	a1=a3*a1
	if(r1-->=0)goto whchton
	a0=ifalt(*r11++)
	return (r14)
	*r6 = a0 = int(a0)
	
/*****************************************************************************/
/*									     */
/*			PROGRAM COEFFICIENTS				     */
/*									     */
lambda:	float 0.981481481
threshs:float 265000.		/* Low band energy threshold */
thresh: float .95		/* c magnitude threshold .9814814 */
minus1:	float -1.
zero:	float 0.
one:	float 1.
two:	float 2.0
three:	float 3.0
four:	float 4.0
_romtbl:int 1, 2, 3, 10
	int 4, 5, 6, 11
	int 7, 8, 9, 12
	int 14, 0, 15, 13

/*	Low bandpass filter coefficients				     */
lcoef:	float .069190871
	float -.981758, .494963, 1.656126378, 1.
	float -.240798, .870132, 0.751487053, 1.

/*	High bandpass filter coefficients				     */
hcoef:	float 0.045015823
	float 1.202186, .557487, -1.470971950, .998564553
	float .660893, .877532, -0.304368084, 0.999705344

/*****************************************************************************/
/*			Angle threshold coefficients			     */
/*									     */
/*			angle = freq. * 180deg. / 2000hz		     */
/*	low band angle computation:  Table value = tan^2(angle) + 1	     */
/*	high band angle computation: Table value = tan^2(180 - angle) + 1    */
/*****************************************************************************/
/*		  Table value	freg(hz) freg offset(%)			     */
/*									     */
/* Wide aperture:							     */
lotbl3:	float	    3.97154     /* 665	-4.6 */
	float	    6.13336    	/* 735	 5.4 */
	float	    6.11845     /* 735	-4.5 */
	float	    11.8574    	/* 812	 5.4 */
	float	    11.7947     /* 812	-4.7 */
	float	    38.8135    	/* 897   5.3 */
	float	    37.6823     /* 895	-4.9 */
	float	    4156.93     /* 990	 5.2 */
hitbl3: float	    6.02902    	/* 1267   4.8 */
	float	    19.2554     /* 1146  -5.2 */
	float	    2.91950    	/* 1398	  4.6 */
	float	    6.12220    	/* 1265  -5.3 */
	float	   1.738967    	/* 1548   4.8 */
	float	    2.88772    	/* 1401  -5.1 */
	float	   1.243161    	/* 1708   4.6 */
	float	   1.749901    	/* 1546  -5.3 */

/* Narrow aperture:							     */
lotblt:	float     4.38518    	/* 683  -2.0 */
	float     5.31677    	/* 714   2.4 */
	float     6.93326	/* 752	-2.3 */
	float     9.33380    	/* 788   2.3 */
	float     14.69761	/* 832  -2.3 */
	float     23.9671    	/* 869   1.9 */
	float     55.7752	/* 914  -2.8 */
	float     322.374    	/* 965   2.6 */
hitblt: float     7.08055    	/* 1245   2.9 */
	float     12.441	/* 1183  -2.1 */
	float     3.32711    	/* 1369   2.5 */
	float     4.59431	/* 1309  -2.0 */
	float     1.916202    	/* 1514   2.5 */
	float     2.397		/* 1447  -2.0 */
	float     1.312574    	/* 1675   2.6 */
	float     1.532391    	/* 1599  -2.1 */

/*****************************************************************************/
/*									     */
/*			PROGRAM STATE VARIABLES				     */
/*									     */

.rsect ".hi_ram"
outl:	float 0.0		/* output from low bandpass filter */
outh:	float 0.0		/* output from high bnadpass filter */
lostate:4*float 0.		/* Low bandpass filter state variables */
histate:4*float 0.		/* High bandpass filter state variables */
det:	float 0.		/* (a1*det)^2 */
aperture:int 0			/* contains present aperture size */
tone:	int 0			/* present low or high band tone detected */
_ttdigit:2*int 0		/* contains past 2 digits detected */

/*****************************************************************************/
/*									     */
/*	Low band state variables (frame by frame basis). 		     */
/*									     */
/*****************************************************************************/
shortel:float 0.	/* energy in low band for one frame */
autoc22l:float 0.	/* R(2,2) */
autoc12l:float 0.	/* R(1,2) */
autoc11l:float 0.	/* R(1,1) */
crcor2l:float 0.	/* R(0,2) */
crcor1l:float 0.	/* R(0,1) */
y2l:	float 0.	/*y(n-2)*/
y1l:	float 0.	/*y(n-1)*/

/*****************************************************************************/
/*									     */
/*	High band state variables (frame by frame basis). 		     */
/*									     */
/*****************************************************************************/
autoc22:float 0.	/* R(2,2) */
autoc12:float 0.	/* R(1,2) */
autoc11:float 0.	/* R(1,1) */
crcor2:	float 0.	/* R(0,2) */
crcor1:float 0.		/* R(0,1) */
y2:	float 0.	/*y(n-2)*/
y1:	float 0.	/*y(n-1)*/
x:	float 0.

/*****************************************************************************/
/*				DMA location				     */
/*****************************************************************************/
bufpnt:	int 0
	.align 4
inbuf:	105*float 0.
ilimit:	float 0.
bufio:	53*float 0.
output:	300*int 0
